freeze;

// THINK:  check that weights on poly ring are *never* used, but that
// instead the sequence of generic weights is deployed when needed
// for calculation of degrees, or whatever.

/////////////////////////////////////////////////////////////////////////
// format.m
/////////////////////////////////////////////////////////////////////////
// The basic abstract format intrinsics.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////
// Philosophy
// ----------
// A format is a KEY VARIETY, an affine Gorenstein variety V, together with a
// choice of (diagonal) C^* action (acting on V, and on the given coordinates
// of its ambient affine space as eigencoordinates). We will need information
// about the graded free resolution of the ideal of V - in many cases, it is
// this data and the weights of the action that are really needed, and the
// equations defining V can often be omitted. The point of V is to create new
// projective varieties by regular pullback.
//
// The archetype key variety is the cone on Gr(2,5), and any variety defined by
// the five maximal Pfaffians of a graded skew 5x5 matrix arises as a regular
// pullback from this key variety.
//
// Creation
// --------
// To create a new format, use the intrinsic "InternalCreateFormat".
//
// Required attributes
// -------------------
// Every format must have the following attributes specified:
// * <MonStgElt>name
//   This is a brief name for the format, intended for printing. It should be
//   chosen to fit into the phrase "Format <name>".
// * <RngIntElt>codimension
//   The codimension of the format - i.e. the codimension of V.
// * <SeqEnum[SeqEnum[RngIntElt]]>betti_data
//   The Betti data of the format. This is a sequence of sequences, the first
//   of which contains the degrees of the equations, the second, the degrees of
//   the first syzygies, and so on; the last (which is the c+1-th, if c is the
//	 codimension) should be a singleton sequence [k], where k is the adjunction
//	 number.
//
// Optional attributes
// -------------------
// The following attributes are optional, but should be set if known:
// * <SeqEnum[RngMPolElt]>key_equations
//   A sequence of polynomials defining the key variety V.
// * <SeqEnum[RngIntElt]>key_weights
//   A sequence of integers defining the ambient weights of the key variety V.
// * <RngUPolElt>hilbert_numerator
//   The Hilbert numerator of V.
// * <Any>numerical_data
//   A junk box attribute containing anything used in creation that would
//   normally be thrown away but is still wanted for printing.
//
// Optional functions
// ------------------
// Optional functions can be set to augment the basic format code. The
// available optional functions are:
// * extra_to_string( <Fmt>F, <MonStgElt>l ) -> MonStgElt
//   Given the format F, should return a string containing any extra
//   information to be displayed when printing F at print level l. The print
//   levels that should be supported are "Minimal", "Default", "Maximal", and
//   "Magma".
// * extra_to_latex( <Fmt>F ) -> MonStgElt
//   Given the format F, should return a LaTeX formatted string containing and
//   extra information about this format.
// * max_weight_non_cone( <Fmt>F ) -> RngIntElt
//   Given the format F, should return the largest ambient weight for any
//   realisation of F that is not a cone.
// * weights_from_equations( <Fmt>F, <RngIntElt>d, <RngIntElt>minwt,
//      <RngIntElt>maxwt ) -> SeqEnum[RngIntElt]
//   A maximal list of possible weights (excluding weight 1) for any
//   realisation of the format F, assuming that
//      (a) the model is not a cone,
//      (b) the orbifold singular locus is of dimension at most 'd',
//      (c) 'minwt' is the smallest allowable weight, and 'maxwt' is the
//          largest.
// * candidate_weights( <Fmt>F, <RngIntElt>l, <RngIntElt>k,
//      <RngIntElt>SingDim, <RngIntElt>minwt, <RngIntElt>maxwt )
//      -> SeqEnum[SeqEnum[RngIntElt]]
//   Should return a sequence of ambient weights W of all possible X in P(W)
//   that could realise the given data:
//       F a format, l = length of W, K_X = O_X(k),
//       dim(orbifold-sing-locus(X)) <= SingDim,
//       Min(W) >= minwt, Max(W) <= maxwt.
//   As usual, it may optimistically return W for X that do not exist, but it
//   should not miss W for X that do exist.
// * analyse_zero_stratum( <Fmt>F, <SeqEnum[RngIntElt]>W, <RngIntElt>r,
//      <RngIntElt>k, <MonStgElt>singtype ) -> BoolElt,SetEnum[FmtSng]
//   Should return true if there is no easy contradiction to the unique ambient
//   index r point being of singularity type singtype, or absent on a
//   realisation of F. If true, should also return a set of singularities of
//   index r that could appear on a numerical variety with canonical class O(k)
//   in format F embedded in weighted projective space P(W). You may assume
//   that there is a unique point in P(W) of index divisible by r.
// * filter( <FmtCnd>X ) -> BoolElt
//   A filter function that will sift through the candidate numerical varieties
//   accepting or rejecting them based on format-specific knowledge. The filter
//   function should accept a candidate X as an argument, and return a boolean
//   (where true = accept, false = reject).
/////////////////////////////////////////////////////////////////////////

import "utilities.m": knapsack;
import "strings.m": split_to_lc, intseq_to_str, weights_to_string;
import "singularities.m": check_properties,singularity_is_of_type;

declare type Fmt;

declare attributes Fmt:
    // Required attributes
    name,                   // Short format name
    codimension,            // Codimension
    betti_data,             // Numerical information about the free resolution
    // Optional attributes
    equations,              // Key variety info
    generic_weights,        // The weights of the variables of the key variety
    hilbert_numerator,      // Hilbert numerator of the key variety
    numerical_data,         // Auxiliary numerical data that the format knows
    // Optional functions
    extra_to_string,        // Format-dependent printing
    extra_to_latex,         // Format-dependent LaTeX printing
    max_weight_non_cone,    // The largest non-cone ambient weight
    weights_from_equations, // The maximal list of possible weights
    candidate_weights,      // Integer sequence of ambient candidate weights
    analyse_zero_stratum,   // User function as above
    filter,                 // User function as above
	// Private attributes
    adjunction_degree;      // Adjunction degree of the key variety

declare verbose Format,1;   // Verbose flag is on (1) or off (0)

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// True iff x is false or a function.
function is_function(x)
    return (Type(x) eq BoolElt and x eq false) or Type(x) eq Program
	or Type(x) eq UserProgram;
		// THINK: Program vs UserProgram? magma version issue?
end function;

// Returns a set of candidate embedding weights W for the format. Here 'l' is
// the length of the weights, 'h' is the sum of the weights, 'weights' is
// the possible weights (with multiplicity, excluding 1's), and 'minweight' is
// the minimum allowed weight.
function possible_weights(l,h,weights,minweight)
    // Initialize the data
    results:={PowerSequence(Integers())|};
    Sort(~weights);
    // We will add in 1's in all ways possible before considering bigger weights
    if minweight gt 1 then
        range:=[0];
        weights:=[Integers() | a : a in weights | a ge minweight];
    else
        range:=[l..Maximum(0,l - #weights) by -1];
    end if;
    for i in range do
        // Create the sequence of 1's
        W1s:=[Integers() | 1 : j in [1..i]];
        // Are there enough weights to add up to h?
        if l - i gt #weights then
            break i;
        end if;
        if l eq i then
            if h eq i then
                Include(~results,W1s);
            end if;
        else
            // Strip off big weights that couldn't appear
            small_sum:=&+[Integers() | weights[k] : k in [1..l - i - 1]] + i;
            temp_wts:=[Integers() | a : a in weights | small_sum + a le h];
            // Perform the knapsack
            poss_wts:=knapsack(temp_wts,h - i,l - i);
            // We special case on the number of 1's to guarantee well-formedness
            if i eq 0 then
                for W in poss_wts do
                    if IsWellformed(W) then
                        Include(~results,Sort(W));
                    end if;
                end for;
            elif i eq 1 then
                for W in poss_wts do
                    if GCD(W) eq 1 then
                        Include(~results,W1s cat Sort(W));
                    end if;
                end for;
            else
                for W in poss_wts do
                    Include(~results,W1s cat Sort(W));
                end for;
            end if;
        end if;
    end for;
    // Return the possible weights
    return results;
end function;

// Generate a set of all possible sings (of given singtype) of index r that
// could live on a dim-dimensional variety in P(W). This can contain too many
// sings, as long as their Porb can be calculated -- we essentially ignore the
// singtype. We also don't attempt to remove r from W since, it might not be
// there (e.g. can have 1/5(1,1,4) on X_30 in P(1,1,4,10,15)).
function possible_singularities(r,W,k,dim,singtype)
    // If necessary, split singtype into a sequence
    if Type(singtype) eq MonStgElt then
        singtype:=split_to_lc(singtype);
    end if;
    // Initialize the results
    sings:={PowerStructure(FmtSng)|};
    // If the singularities are isolated, we can prune the weights
	if "isolated" in singtype then
		W:=[Integers() | a : a in W | GCD(r,a) eq 1];
	end if;
	// Is there anything to do?
	if dim gt #W then return sings; end if;
	// We only know how to handle quotient singularities
	error if not "quotient" in singtype,
	   "Cannot handle singularities that are not quotient singularities";
    // Seek 1/r(a1,a2,..a_dim) such that sum(ai) + k = i*r for some i.
    // (Biggest poss i is (sum(ai) + k) / r, where a1 >= a2, etc.)
    Sort(~W);
    smallest_i:=Ceiling((&+W[1..dim] + k) / r);
    Reverse(~W);
    biggest_i:=Floor((&+W[1..dim] + k) / r);
    for i in [smallest_i..biggest_i] do
        for A in knapsack(W,i * r - k,dim) do
            P:=Singularity(r,A,k);
            if singularity_is_of_type(P,singtype) then
                Include(~sings,P);
            end if;
        end for;
    end for;
    // Return the possible singularities
    return sings;
end function;

// The largest ambient weight for any realisation of F that is not a cone.
function max_weight_non_cone(F)
    return assigned F`max_weight_non_cone select F`max_weight_non_cone(F)
                                            else Maximum(GenericWeights(F));
end function;

// For format F, integer d (dimension) and integer w (weight) return the
// maximum number of occurrences of w there could be in weights W for
// a realisation of F in P(W) with singular locus of dim <= d.
// (Called by weights_from_equations below.)
// For notation, suppose a proposed variable of weight w is called x.
function max_number(F,d,w)
    E:=Equations(F);
    R:=Universe(E);
    wts:=GenericWeights(F); // Integers, but not necessarily all > 0.
    divble_vars:=[R|];      // The ambient vars for which could pullback as x^?
    for i in [1..#wts] do
        // NB. If wt[i] = 0, that's fine - if < 0 then doesn't appear in eqns
        if wts[i] gt 0 and (wts[i] mod w) eq 0 then
            Append(~divble_vars,R.i);
        else
            Append(~divble_vars,R ! 0);
        end if;
    end for;
    // Figure out the eqns that could have a pure x^? term after pullback.
    divble_eqns:=[Evaluate(f,divble_vars) : f in E];
    divble_eqns:=[R | f : f in divble_eqns | f ne 0];
    // Note that if divble eqns are independent, then you could have this many
    // vars of wt w; otherwise fewer, so this is a safe upper bound.
    return d + 1 + #divble_eqns;
end function;

// A maximal list of possible weights (excluding weight 1) for any realisation
// of the format F, assuming that (a) the model is not a cone, (b) the orbifold
// singular locus is of dimension at most 'd', and (c) 'minwt' is the smallest
// allowable weight, and 'maxwt' is the largest.
function weights_from_equations(F,d,minwt,maxwt)
    if assigned F`weights_from_equations then
        return F`weights_from_equations(F,d,minwt,maxwt);
    else
        wts:=[Integers()|];
        for w in [maxwt..Maximum(minwt,2) by -1] do
            wts cat:= [Integers() | w : i in [1..max_number(F,d,w)]];
        end for;
        return wts;
    end if;
end function;

/////////////////////////////////////////////////////////////////////////
// Creation intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic InternalCreateFormat(
    name::MonStgElt,
    codim::RngIntElt,
    betti::SeqEnum[SeqEnum[RngIntElt]] :
    key_equations:=false,
    key_weights:=false,
    hilbert_numerator:=false,
    extra_to_string:=false,
    extra_to_latex:=false,
    max_weight_non_cone:=false,
    weights_from_equations:=false,
    candidate_weights:=false,
    analyse_zero_stratum:=false,
    filter:=false ) -> Fmt
{Create a new format with the given data}
    // Sanity check
    name:=Trim(name);
    require #name gt 0: "The format name must not be empty";
    require codim ge 0: "The codimension must be a non-negative integer";
    require #betti ne 0: "The Betti data cannot be empty";
    require (key_equations cmpeq false) or (Type(key_equations) eq SeqEnum and Type(Universe(key_equations)) eq RngMPol):
        "Parameter 'key_equations' must be a sequence of polynomials";
    require (key_weights cmpeq false) or (Type(key_weights) eq SeqEnum and Universe(key_weights) cmpeq Integers()):
        "Parameter 'key_weights' must be a sequence of integers";
    require (hilbert_numerator cmpeq false) or (Type(hilbert_numerator) eq RngUPolElt):
        "Parameter 'hilbert_numerator' must be a univariate polynomial";
    require is_function(extra_to_string):
        "Parameter 'extra_to_string' must be a function";
    require is_function(extra_to_latex):
        "Parameter 'extra_to_latex' must be a function";
    require is_function(max_weight_non_cone):
        "Parameter 'max_weight_non_cone' must be a function";
    require is_function(weights_from_equations):
        "Parameter 'weights_from_equations' must be a function";
    require is_function(candidate_weights):
        "Parameter 'candidate_weights' must be a function";
    require is_function(analyse_zero_stratum):
        "Parameter 'analyse_zero_stratum' must be a function";
    require is_function(filter):
        "Parameter 'filter' must be a function";
    // Create the format object
    F:=New(Fmt);
    // Assign the required data
    F`name:=name;
    F`codimension:=codim;
    F`betti_data:=betti;
    // Assign the optional functions
    if not key_equations cmpeq false then
        F`equations:=key_equations;
    end if;
    if not key_weights cmpeq false then
        F`generic_weights:=key_weights;
    end if;
    if not hilbert_numerator cmpeq false then
        F`hilbert_numerator:=hilbert_numerator;
    end if;
    if not extra_to_string cmpeq false then
        F`extra_to_string:=extra_to_string;
    end if;
    if not extra_to_latex cmpeq false then
        F`extra_to_latex:=extra_to_latex;
    end if;
    if not max_weight_non_cone cmpeq false then
        F`max_weight_non_cone:=max_weight_non_cone;
    end if;
    if not weights_from_equations cmpeq false then
        F`weights_from_equations:=weights_from_equations;
    end if;
    if not candidate_weights cmpeq false then
        F`candidate_weights:=candidate_weights;
    end if;
    if not analyse_zero_stratum cmpeq false then
        F`analyse_zero_stratum:=analyse_zero_stratum;
    end if;
    if not filter cmpeq false then
        F`filter:=filter;
    end if;
    // Return the format object
    return F;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// Display intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Print(F::Fmt,l::MonStgElt)
{Print the format F at print level l}
    if l ne "Minimal" and l ne "Default" and l ne "Maximal" and l ne "Magma" and
       l ne "Latex" then l:="Default"; end if;
    printf Trim(FormatToString(F : level:=l));
end intrinsic;

intrinsic FormatToString(F::Fmt : level:="Default") -> MonStgElt
{A string representation of the format F}
    // Sanity check
    require Type(level) eq MonStgElt and (level eq "Minimal" or
        level eq "Default" or level eq "Maximal" or level eq "Magma" or
        level eq "Latex"):
        "Parameter 'level' should be one of \"Minimal\", \"Default\", \"Maximal\", \"Magma\", or \"Latex\"";
    // Is this the LaTeX print level?
    if level eq "Latex" then
        try
            str:="X_{" cat weights_to_string(EquationDegrees(F) :
                    brackets := "", latex := true) cat "}&" cat
                    IntegerToString(Codimension(F));
        catch e;
            str:="X&" cat IntegerToString(Codimension(F));
        end try;
        if assigned F`extra_to_latex then
            try
                s:=F`extra_to_latex(F);
            catch e;
                s:="";
            end try;
            while #s ne 0 and s[#s] eq "\n" do
                s:=s[1..#s - 1];
            end while;
            if #s ne 0 then
                str cat:= "&" cat s;
            end if;
        end if;
        return str cat "\\\\\n";
    end if;
    // Is this the Magma print level?
    if level eq "Magma" then
        if assigned F`extra_to_string then
            try
                str:=F`extra_to_string(F,level);
            catch e;
                str:="";
            end try;
            if #str ne 0 then return str; end if;
        end if;
        level:="Minimal";
    end if;
    // We begin with the basic data
    str:=Sprintf("Format %o of codimension %o.\n",Name(F),Codimension(F));
    // If this isn't the minimal print level then we extend.
    if level ne "Minimal" then
        try
            adjdeg:=AdjunctionDegree(F);
            str cat:= Sprintf("Adjunction degree: %o\n",adjdeg);
        catch e;
        end try;
        try
            eqwts:=EquationDegrees(F);
            if #eqwts ne 0 then
                str cat:= Sprintf("Equation degrees: %o\n",
                                                          intseq_to_str(eqwts));
            end if;
        catch e;
        end try;
        if assigned F`extra_to_string then
            try
                s:=F`extra_to_string(F,level);
            catch e;
                s:="";
            end try;
            while #s ne 0 and s[#s] eq "\n" do
                s:=s[1..#s - 1];
            end while;
            if #s ne 0 then
                str cat:= s cat "\n";
            end if;
        end if;
    end if;
    // Return the string
    return str;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// Access intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Name(F::Fmt) -> MonStgElt
{The format name}
    return F`name;
end intrinsic;

intrinsic BettiData(F::Fmt) -> SeqEnum[SeqEnum[RngIntElt]]
{The Betti data of the format F}
    require assigned F`betti_data: "Betti data not assigned";
    return F`betti_data;
end intrinsic;

intrinsic Dimension(F::Fmt,W::SeqEnum) -> RngIntElt
{The dimension of the format F when embedded in wps with weights W}
    require assigned F`codimension: "Codimension not assigned";
    return #W - 1 - F`codimension;
end intrinsic;

intrinsic Codimension(F::Fmt) -> RngIntElt
{The generic codimension of the format F}
    require assigned F`codimension: "Codimension not assigned";
    return F`codimension;
end intrinsic;

intrinsic GenericWeights(F::Fmt) -> SeqEnum[RngIntElt]
{The ambient weights of the format F}
    if not assigned F`generic_weights then
        require assigned F`equations:
            "Generic weights cannot be recovered (equations not assigned)";
        R:=Universe(F`equations);
        F`generic_weights:=VariableWeights(R);
    end if;
    return F`generic_weights;
end intrinsic;

intrinsic Equations(F::Fmt) -> SeqEnum[RngMPolElt]
{The equations of the format F}
    require assigned F`equations: "Equations not assigned";
    return F`equations;
end intrinsic;

intrinsic EquationDegrees(F::Fmt) -> SeqEnum[RngIntElt]
{The degrees of the equations of the format F}
    require assigned F`betti_data: "Betti data not assigned";
    return #F`betti_data ge 2 select F`betti_data[2] else [Integers()|];
end intrinsic;

intrinsic SyzygyDegrees(F::Fmt,n::RngIntElt) -> SeqEnum[RngIntElt]
{The degrees of the n-th syzygies of the format F}
    require assigned F`betti_data: "Betti data not assigned";
    return #F`betti_data ge n + 2 select F`betti_data[n + 2] else [Integers()|];
end intrinsic;

// If X in wps P = P(W) is of format F then the adjunction number k (which is
// only a function of the format, not of W) computes the canonical sheaf of
// X, w_X, as follows:  w_X = w_P(k).
intrinsic AdjunctionDegree(F::Fmt) -> RngIntElt
{The adjunction degree of the format F (if Gorenstein)}
    if not assigned F`adjunction_degree then
        require assigned F`codimension: "Codimension not assigned";
        require assigned F`betti_data: "Betti data not assigned";
        endofres:=F`codimension + 1;
        require IsDefined(F`betti_data,endofres):
            "Adjunction degree cannot be recovered";
        k:=F`betti_data[endofres];
		require #k eq 1: "Format is not Gorenstein";
		F`adjunction_degree:=k[1];
    end if;
    return F`adjunction_degree;
end intrinsic;

intrinsic HilbertNumerator(F::Fmt) -> RngUPolElt
{The Hilbert numerator of the format F}
    if not assigned F`hilbert_numerator then
        require assigned F`betti_data: "Betti data not assigned";
        R:=PolynomialRing(Rationals());
        t:=R.1;
        num:=[[Integers() | a : a in n] : n in F`betti_data];
         f:=&+[R | (-1)^(i-1) * &+[R | t^j : j in num[i]] : i in [1..#num]];
	// GDB THINK: should allow numerator to be Laurent polynomial - it's
	// allowed in [BKZ12].
        F`hilbert_numerator:=f;
	end if;
    return F`hilbert_numerator;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// Generic format analysis intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic CandidateWeights(F::Fmt,l::RngIntElt,k::RngIntElt,a::RngIntElt :
    SingDim:=0, minweight:=1) -> SeqEnum[SeqEnum[RngIntElt]]
{A sequence of weights (bounded above by a) of length l for candidate numerical varieties X in format F and with canonical class K_X = O_X(k); in particular each weight is an integer sequence of length l.}
    // Note: dim(X) = n, where l = n + Codimension(F) + 1.
    // Sanity check
    bool,SingDim:=IsCoercible(Integers(),SingDim);
    require bool and SingDim ge 0:
        "Parameter 'SingDim' must be a non-negative integer";
    bool,minweight:=IsCoercible(Integers(),minweight);
    require bool and minweight ge 1:
        "Parameter 'minweight' must be a positive integer";
    // Does a custom function exist? If so, use it, otherwise fall back on the
    // default method.
    if assigned F`candidate_weights then
        return F`candidate_weights(F,l,k,SingDim,minweight,a);
    else
    	weights:=weights_from_equations(F,SingDim,minweight,a);
    	h:=AdjunctionDegree(F) - k;
    	return possible_weights(l,h,weights,minweight);
    end if;
end intrinsic;

intrinsic AnalyseZeroStratum(F::Fmt,W::SeqEnum[RngIntElt],r::RngIntElt,
    k::RngIntElt : singtype:="isolated,quotient,canonical")
    -> BoolElt,SetEnum[FmtSng]
{True if there is no easy contradiction to the unique ambient index r point being of singularity type singtype (default: "isolated,quotient,canonical"), or absent on a realisation of F. If true, also gives a set of singularities of index r that could appear on a numerical variety with canonical class O(k) in format F embedded in weighted projective space P(W). Assumes that there is a unique point in P(W) of index divisible by r.}
    // Sanity check
    require Type(singtype) eq MonStgElt:
        "Parameter 'singtype' must be a string";
    singtype:=split_to_lc(singtype);
    bool,msg:=check_properties(singtype);
    require bool: msg;
    // Does a custom function exist? If so, use it, otherwise fall back on the
    // default method.
    if assigned F`analyse_zero_stratum then
        bool,sings:=F`analyse_zero_stratum(F,W,r,k,Join(Setseq(singtype),","));
        if bool then
            return true,sings;
        else
            return false,_;
        end if;
    else
    	sings:=possible_singularities(r,W,k,#W - 1 - Codimension(F),singtype);
    	return true,sings;
    end if;
end intrinsic;

intrinsic FormatIsVerbose() -> BoolElt
{True iff the verbose flag 'Format' is nonzero}
    return IsVerbose("Format");
end intrinsic;

intrinsic FormatSetVerbose(v::BoolElt)
{Set the verbose flag 'Format' to off if v is false and on if v is true}
    SetVerbose("Format",v);
end intrinsic;
