freeze;

/////////////////////////////////////////////////////////////////////////
// rr.m
/////////////////////////////////////////////////////////////////////////
// Riemann-Roch calculations for candidate numerical varieties.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// Q is a seq of numbers, n a positive integer. Returns the seq [Q[i] - Q[i-1]]
// repeated n times.
function recursive_difference(Q,n)
	for j in [1..n] do
		Q:=[Universe(Q) | Q[i] - Q[i - 1] : i in [2..#Q]];
	end for;
	return Q;
end function;

// h a hilbert series (as a rational function) of a variety of dimension d, and
// B a basket (to determine a trunctation of h that is regular, i.e. has no
// contributions coming from B).
// Note: ignore constant term, since there is no vanishing for h^top.
function hilb_deg(h,d,B)
    r:=LCM([Integers() | Index(p) : p in B]);
    S:=PowerSeriesRing(CoefficientRing(Parent(h)),(d+1) * r + 1);
	coeffs:=Coefficients(S ! h);
	truncated_coeffs:=[Universe(coeffs) | coeffs[1 + r * i] : i in [1..d+1]];
	// Use differencing to 'differentiate' h to find leading term
	return recursive_difference(truncated_coeffs,d)[1] / r^d;
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

intrinsic Degree(X::FmtCnd,B::SeqEnum[FmtSng]) -> FldRatElt
{}
    // Sanity check
    d:=Dimension(X);
    require &and[Dimension(Q) le d : Q in B]:
        Sprintf("The singularities must be of dimension at most %o",d);
    // Compute the degree w.r.t. the basket B
    return hilb_deg(HilbertSeries(X),d,B);
end intrinsic;

intrinsic Degree(X::FmtCnd) -> FldRatElt
{The degree of the candidate numerical variety variety X}
    // Sanity check
	Bs:=Baskets(X);
	require #Bs eq 1:
	   "There is more than one basket: You must specify which basket to use.";
    B:=Representative(Bs);
    d:=Dimension(X);
    require &and[Dimension(Q) le d : Q in B]:
        Sprintf("The singularities must be of dimension at most %o",d);
    // Compute the degree w.r.t. the basket B
    return hilb_deg(HilbertSeries(X),d,B);
end intrinsic;
