freeze;

/////////////////////////////////////////////////////////////////////////
// utilities.m
/////////////////////////////////////////////////////////////////////////
// A collection of useful utility functions and intrinsics.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

import "singularities.m": sing_cache_clear;
import "candidate.m": cand_cache_clear;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// weighted_degree
/////////////////////////////////////////////////////////////////////////
// Input: Polynomial f, sequence of weights W (any Z-module elements).
// Output: Degree of f with respect to the weights W.
/////////////////////////////////////////////////////////////////////////

// Weighted degree of poly f w.r.t wts W
function weighted_degree(f,W)
    // Sanity check
    R:=Parent(f);
    error if not #W eq Rank(R),
        "The number of weights in W must equal the number of variables in the polynomial f";
    // Compute the weighted degrees of the terms
    term_degs:={&+[Degree(m,R.i) * W[i] : i in [1..Rank(R)]] : m in Terms(f)};
    // Assert that there is ony one degree and return it
    error if not #term_degs eq 1,
        "Polynomial f is not homogeneous with respect to weights W";
    return Representative(term_degs);
end function;

/////////////////////////////////////////////////////////////////////////
// knapsack_process
/////////////////////////////////////////////////////////////////////////
// Input: Q is a sequence of integers.
// Output: A Process generating all subsequences of Q of length k whose
// sum is n. Each subsequence is sorted in reverse numerical order.
/////////////////////////////////////////////////////////////////////////

// Given a current index I, returns true followed by the next index in the
// sequence, or false if there are no more indices.
function next_index(I,k,Q)
    i:=#I;
    localmax:=#Q - k + i;
    if i eq 0 then
        return true,[1..k];
    end if;
    q:=Q[I[i]];
    I[i] +:= 1;
    while I[i] le localmax and Q[I[i]] eq q do
        I[i] +:= 1;
    end while;
    if I[i] gt localmax then
        if i eq 1 then
            return false,_;
        else
            bool,I:=$$(I[1..i - 1],k,Q);
            if bool then
                return true,I;
            else
                return false,_;
            end if;
        end if;
    end if;
    J:=[Integers() | I[i] + j : j in [1..k - i]];
    return true,I cat J;
end function;

// Given the indices of a current knapsack solution I, returns true followed by
// the indices of the next solution, or false if there are no more solutions.
function ks_next(I,n,k,Q)
    if #Q lt k then return false,_; end if;
    bool,I:=next_index(I,k,Q);
    while bool and &+[Integers() | Q[i] : i in I] ne n do
        bool,I:=next_index(I,k,Q);
    end while;
    if bool then
        return true,I;
    else
        return false,_;
    end if;
end function;

// Advances the knapsack process to the next solution (if any).
procedure ks_advance(~info)
    error if info[4], "The process has finished.";
    bool,I:=ks_next(info[5],info[1],info[2],info[3]);
    if bool then
        info[5]:=I;
        info[6] +:= 1;
    else
        info[4]:=true;
    end if;
end procedure;

// Returns the current knapsack solution.
function ks_current(info)
    error if info[4], "The process has finished.";
    Q:=info[3];
    return [Universe(Q) | Q[i] : i in info[5]];
end function;

// Returns the current knapsack label.
function ks_label(info)
    error if info[4], "The process has finished.";
    return info[6];
end function;

// True iff the knapsack process has finished.
function ks_is_empty(info)
    return info[4];
end function;

function knapsack_process(Q,n,k)
    Q:=Reverse(Sort(Q));
    bool,I:=ks_next([Integers()|],n,k,Q);
    info:=bool select <n,k,Q,false,I,1> else <n,k,Q,true>;
    return InternalCreateProcess("Knapsack",info,ks_is_empty,ks_advance,
                                                           ks_current,ks_label);
end function;

/////////////////////////////////////////////////////////////////////////
// knapsack
/////////////////////////////////////////////////////////////////////////
// Input: Q is a sequence of integers.
// Output: Set of all subsequences of Q of length k whose sum is n.
// Each subsequence is sorted in reverse numerical order.
/////////////////////////////////////////////////////////////////////////
/*
   Q:=[1,2,3,4,5,6];
   knapsack(Q,9,3);
   {
      [ 6, 2, 1 ],
      [ 4, 3, 2 ],
      [ 5, 3, 1 ]
   }
*/

function knapsack_recurse(Q,Qdist,n,k)
    results:={PowerSequence(Integers())|};
    if k ne 1 and #Q ge k then
        smallest_sum:=&+[Integers() | Q[i] : i in [#Q - k + 2..#Q]];
        greatest_sum:=&+[Integers() | Q[i] : i in [1..k - 1]];
        Q:=[Integers() | a : a in Q | (a + smallest_sum le n) and
                                                       (a + greatest_sum ge n)];
        if #Q gt k then
            Qdist:=[Integers() | a : a in Qdist | a in Q];
            for a in Qdist do
                Remove(~Q,Index(Q,a));
                partial:=$$(Q,Qdist,n - a,k - 1);
                Q:=[Integers() | b : b in Q | b ne a];
                for Wini in partial do
                    Include(~results,[Integers() | a] cat Wini);
                end for;
            end for;
        elif #Q eq k and &+Q eq n then
            Include(~results,Q);
        end if;
    elif k eq 1 and n in Q then
        Include(~results,[Integers() | n]);
    end if;
    return results;
end function;

function knapsack(Q,n,k)
    Q:=Reverse(Sort(Q));
    Qdist:=Reverse(Sort(Setseq(Seqset(Q))));
    return knapsack_recurse(Q,Qdist,n,k);
end function;

//////////////////////////////////////////////////////////////////////
// knapsack_with_repeats
//////////////////////////////////////////////////////////////////////
// Input: A sequence of integers Q.
// Output: A set whose elements are sequences of elements from Q
// (allowing repeats) of length k such that their sum is n. Each
// sequence is sorted in reverse numerical order.
//////////////////////////////////////////////////////////////////////
/*
   Q:=[-1,0,1,2,3,4];
   n:=5;
   k:=3;
   knapsack_with_repeats(Q,n,k);
   {
       [ 3, 3, -1 ],
       [ 3, 1, 1 ],
       [ 2, 2, 1 ],
       [ 3, 2, 0 ],
       [ 4, 1, 0 ],
       [ 4, 2, -1 ]
   }
*/

function knapsack_with_repeats_recurse(Q,n,k)
    results:={PowerSequence(Integers())|};
    if k ne 1 and #Q gt 0 then
        smallest_sum:=Q[#Q] * (k-1);
        Q:=[a : a in Q | a + smallest_sum le n];
        newQ:=Q;
        for a in Q do
            partial := $$(newQ,n-a,k-1);
            for Wini in partial do
                Include(~results,[a] cat Wini);
            end for;
            Remove(~newQ,1);
        end for;
    elif k eq 1 and n in Q then
        Include(~results,[n]);
    end if;
    return results;
end function;

function knapsack_with_repeats(Q,n,k)
        Q:=Reverse(Sort(Setseq(Seqset(Q))));
        return knapsack_with_repeats_recurse(Q,n,k);
end function;

/////////////////////////////////////////////////////////////////////////
// toric_knapsack
/////////////////////////////////////////////////////////////////////////
// Input: Q is a sequence of sequences of integers of the same length,
// and w is another sequence of integers of that same length.
// Ouput: The polyhedron in the positive orthant whose points are the
// coeffients of linear relations in Q that achieve w.
/////////////////////////////////////////////////////////////////////////
/*
    Hs:=[ PowerSequence(Integers()) |
            [ 1, 1, 2, 1, 1, 1, 3, 1, 3, 1 ],
            [ 3, 4, 7, 4, 4, 4, 10, 4, 11, 4 ],
            [ 7, 9, 16, 10, 10, 10, 23, 10, 25, 10 ],
            [ 13, 17, 31, 20, 20, 20, 44, 20, 48, 20 ],
            [ 22, 29, 52, 34, 35, 35, 74, 35, 81, 35 ],
            [ 34, 45, 81, 53, 55, 56, 116, 56, 126, 56 ],
            [ 50, 66, 119, 78, 81, 83, 170, 84, 186, 84 ],
            [ 70, 93, 167, 110, 114, 117, 239, 119, 261, 120 ],
            [ 95, 126, 227, 150, 155, 159, 324, 162, 354, 164 ],
            [ 125, 166, 299, 198, 205, 210, 427, 214, 467, 217 ],
            [ 161, 214, 385, 255, 265, 271, 550, 276, 601, 280 ],
            [ 203, 270, 486, 322, 335, 343, 694, 349, 759, 354 ],
            [ 252, 335, 603, 400, 416, 427, 862, 434, 942, 440 ],
            [ 308, 410, 738, 490, 509, 523, 1054, 532, 1152, 539 ],
            [ 372, 495, 891, 592, 615, 632, 1273, 644, 1392, 652 ],
            [ 444, 591, 1064, 707, 735, 755, 1520, 770, 1662, 780 ],
            [ 525, 699, 1258, 836, 870, 893, 1797, 911, 1965, 924 ],
            [ 615, 819, 1474, 980, 1020, 1047, 2106, 1068, 2303, 1084 ],
            [ 715, 952, 1714, 1140, 1186, 1218, 2448, 1242, 2677, 1261 ],
            [ 825, 1099, 1978, 1316, 1369, 1407, 2826, 1434, 3090, 1456 ],
            [ 946, 1260, 2268, 1509, 1570, 1614, 3240, 1645, 3543, 1670 ] ];
    cs:=[ Integers() | 3, 12, 28, 54, 93, 146, 216, 306, 416, 549, 708, 894,
            1110, 1359, 1642, 1962, 2322, 2722, 3165, 3654, 4190 ];
    time hyperplane_knapsack(Hs,cs);
    Output:
    0-dimensional polytope with one vertex:
        (0, 2, 0, 0, 0, 0, 0, 0, 0, 1)
    Time: 0.310
*/

function hyperplane_knapsack(Hs,cs)
    // Sanity checks have been done by toric_knapsack below.
    d:=#Hs[1];
    // Move to the graded lattice
    D:=GradedToricLattice(Dual(Parent(LatticeVector(Hs[1]))));
    Hs:=[D | [-cs[i]] cat Hs[i] : i in [1..#Hs]];
    // We start with the positive orthant
    C:=ConeWithInequalities([D.i : i in [2..d + 1]]);
    // Intersect with the hyperplanes as necessary
    for H in Hs do
        if &or[H * r ne 0 : r in MinimalRGenerators(C : algorithm:="FM")] then
            C:=ConeWithInequalities(Inequalities(C) cat [H,-H]);
        end if;
    end for;
    // Finally, return the polyhedron
    return Polyhedron(C);
end function;

function toric_knapsack(Q,w)
    // Get the degenerate case out of the way
	if #Q eq 0 then return EmptyPolyhedron(#w); end if;
	// Sanity check
	d:=#w;
	error if d eq 0,
	   "Argument 2 must be a nonempty integer sequence";
	error if not &and[#H eq d : H in Q],
		"Argument 1 must be a sequence of integer sequences, each having the same length as argument 2";
	bool,Q:=CanChangeUniverse(Q,PowerSequence(Integers()));
	error if not bool,
	   "Argument 1 must be a sequence of integer sequences, each having the same length as argument 2";
    bool,w:=CanChangeUniverse(w,Integers());
    error if not bool,
        "Argument 2 must be a nonempty integer sequence";
    // OK, let the hyperplane knapsack do the work
	return hyperplane_knapsack(RowSequence(Transpose(Matrix(Q))),w);
end function;

/////////////////////////////////////////////////////////////////////////
// Intrinsics
/////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
// IsHalfInteger
/////////////////////////////////////////////////////////////////////////
// Input: A rational number n.
// Output: True iff n = p/2 for an odd integer p.
/////////////////////////////////////////////////////////////////////////

intrinsic IsHalfInteger( n::FldRatElt ) -> BoolElt
{True iff n = p/2 for some odd integer p}
    bool,p:=IsCoercible(Integers(),2 * n);
    return bool select IsOdd(p) else false;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// HalfIntegerCeiling
/////////////////////////////////////////////////////////////////////////
// Input: A rational number n.
// Output: p/2, where p is the smallest odd integer such that n <= p/2.
/////////////////////////////////////////////////////////////////////////

intrinsic HalfIntegerCeiling( n::FldRatElt ) -> FldRatElt
{The smallest value p/2 such that p is an odd integer and n <= p/2}
    c:=Ceiling(2 * n);
    return IsEven(c) select (c + 1) / 2 else c / 2;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// HalfIntegerFloor
/////////////////////////////////////////////////////////////////////////
// Input: A rational number n.
// Output: p/2, where p is the largest odd integer such that p/2 <= n.
/////////////////////////////////////////////////////////////////////////

intrinsic HalfIntegerFloor( n::FldRatElt ) -> FldRatElt
{The largest value p/2 such that p is an odd integer and p/2 <= n}
    c:=Floor(2 * n);
    return IsEven(c) select (c - 1) / 2 else c / 2;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// Coefficient
/////////////////////////////////////////////////////////////////////////
// Input: A rational function in one indeterminant, f, and a positive
// integer n.
// Output: The coefficient of t^n in the expansion of f.
/////////////////////////////////////////////////////////////////////////

intrinsic Coefficient( f::FldFunRatUElt, n::RngIntElt ) -> RngElt
{The coefficient of t^n in the expansion of f}
    require n ge 0: "Argument 2 must be non-negative";
    S:=PowerSeriesRing(CoefficientRing(Parent(f)),n + 1);
    return Coefficient(S ! f, n);
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// FirstNCoefficients
/////////////////////////////////////////////////////////////////////////
// Input: A rational function in one indeterminant, f, and a positive
// integer N.
// Output: A list of length n containing the coefficients of
// t^0,...,t^(N-1) in the expansion of f. The sequence will be padded by
// zeros if appropriate.
/////////////////////////////////////////////////////////////////////////
/*
    K<t> := RationalFunctionField(Rationals());
    f := 1/(1-t)^3;
    FirstNCoefficients(f,15);
    [ 1, 3, 6, 10, 15, 21, 28, 36, 45, 55, 66, 78, 91, 105, 120 ]
*/

intrinsic FirstNCoefficients( f::FldFunRatUElt, N::RngIntElt ) -> SeqEnum
{The coefficients of t^0,...,t^(N-1) in the expansion of f, given as a sequence. The sequence will be padded by zeros if appropriate.}
    // Sanity check
    require N ge 0: "Argument 2 must be non-negative";
    // Recover the variable and coefficient ring
    t:=Parent(Denominator(f)).1;
    K:=CoefficientRing(Parent(f));
    // A small fix in case f has negative valuation
    negvalf:=0;
    while IsDivisibleBy(Denominator(f),t^negvalf) do
        negvalf +:= 1;
    end while;
    negvalf -:= 1;
    S:=PowerSeriesRing(K,N + negvalf);
    Sf:=FieldOfFractions(S) ! f;
    coeffs,firstdeg,stepdeg:=Coefficients(Sf);
    assert stepdeg eq 1;
    if firstdeg ge 0 then
        coeffs:=ZeroSequence(K,firstdeg) cat coeffs;
    else
        coeffs:=coeffs[negvalf..#coeffs];
    end if;
    // Pad with trailing zeros in case power series was a poly, or similar.
    return coeffs cat ZeroSequence(K,N - #coeffs);
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// CoefficientsInRange
/////////////////////////////////////////////////////////////////////////
// Input: A rational function f in one indeterminate and two integers
// n1 <= n2.
// Output: A list of length n2 - n1 + 1 containing the coefficients of
// t^n1,...,t^n2 in the expansion of f. The list will be padded by 0s
// if appropriate.
/////////////////////////////////////////////////////////////////////////

intrinsic CoefficientsInRange( f::FldFunRatUElt, n1::RngIntElt, n2::RngIntElt )
    -> SeqEnum
{A sequence containing the coefficients of t^n1,...,t^n2 in the Laurent expansion of f at zero. The sequence will be padded by zeros if appropriate.}
    // Sanity check
    require n1 le n2: "Argument 2 must be less than or equal to argument 3";
    // Collect the data we need
	R:=Parent(Numerator(f));
	t:=R.1;
	K:=CoefficientRing(R);
	// Is there anything to do?
	if f eq (R ! 0) then return ZeroSequence(K,n2 - n1 + 1); end if;
	// First find any t factors in the denominator - remove and remember.
	negvalf:=0;
	while IsDivisibleBy(Denominator(f),t^negvalf) do
        negvalf +:= 1;
	end while;
	negvalf -:= 1;
	// So t^(-negvalf) is the leading term of f as a Laurent series.
    // If n2 < negvalf, then we already know the answer.
    if n2 lt negvalf then
        return ZeroSequence(K,n2 - n1 + 1);
    // If n1 < negvalf, then the first -negvalf - n1 coeffs are 0.
    else
        initial_zeros:=ZeroSequence(K,Maximum(0,-n1 - negvalf));
        n1:=Maximum([n1,-negvalf]);
        coeffs:=FirstNCoefficients( f * t^negvalf, n2 + 1 + negvalf );
        return initial_zeros cat coeffs[n1 + 1 + negvalf..n2 + 1 + negvalf];
    end if;
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// ComplementaryWeights
/////////////////////////////////////////////////////////////////////////
// Input: A is a 5x5 or 7x7 matrix, (i,j) is the index of an entry.
// Output: The three/ten entries that multipy A[i,j] in the pfaffians.
/////////////////////////////////////////////////////////////////////////

function complementary_entries_5(i,j)
    Q:=[1..5];
    Remove(~Q,Index(Q,i));
    Remove(~Q,Index(Q,j));
    return [[Q[1],Q[2]], [Q[1],Q[3]], [Q[2],Q[3]]];
end function;

function complementary_entries_7(i,j)
    Q:=[1..7];
    Remove(~Q,Index(Q,i));
    Remove(~Q,Index(Q,j));
    return [[Q[1],Q[2]], [Q[1],Q[3]], [Q[1],Q[4]], [Q[1],Q[5]],
                         [Q[2],Q[3]], [Q[2],Q[4]], [Q[2],Q[5]],
                                      [Q[3],Q[4]], [Q[3],Q[5]],
                                                   [Q[4],Q[5]]];
end function;

intrinsic ComplementaryWeights(A::Mtrx[FldRat],i::RngIntElt,j::RngIntElt)
    -> SeqEnum[FldRatElt]
{}
    if NumberOfRows(A) eq 5 and NumberOfColumns(A) eq 5 then
        require i ge 1 and i le 5: "Argument 2 must be in the range 1..5";
        require j ge 1 and j le 5: "Argument 3 must be in the range 1..5";
        require i ne j: "Arguments 2 and 3 cannot be equal";
        return i lt j select [A[w[1],w[2]] : w in complementary_entries_5(i,j)]
                        else [A[w[2],w[1]] : w in complementary_entries_5(i,j)];
    elif NumberOfRows(A) eq 7 and NumberOfColumns(A) eq 7 then
        require i ge 1 and i le 7: "Argument 2 must be in the range 1..7";
        require j ge 1 and j le 7: "Argument 3 must be in the range 1..7";
        require i ne j: "Arguments 2 and 3 cannot be equal";
        return i lt j select [A[w[1],w[2]] : w in complementary_entries_7(i,j)]
                        else [A[w[2],w[1]] : w in complementary_entries_7(i,j)];
    end if;
    // Haven't landed in one of the cases we cover, so raise an error.
    require false: "Argument 1 must be a 5x5 or 7x7 matrix";
end intrinsic;

intrinsic ComplementaryWeights(A::Mtrx[RngInt],i::RngIntElt,j::RngIntElt)
    -> SeqEnum[RngIntElt]
{The sequence of entries that multiply A[i,j] in the pfaffians, where A is a 5x5 or 7x7 matrix.}
    if NumberOfRows(A) eq 5 and NumberOfColumns(A) eq 5 then
        require i ge 1 and i le 5: "Argument 2 must be in the range 1..5";
        require j ge 1 and j le 5: "Argument 3 must be in the range 1..5";
        require i ne j: "Arguments 2 and 3 cannot be equal";
        return i lt j select [A[w[1],w[2]] : w in complementary_entries_5(i,j)]
                        else [A[w[2],w[1]] : w in complementary_entries_5(i,j)];
    elif NumberOfRows(A) eq 7 and NumberOfColumns(A) eq 7 then
        require i ge 1 and i le 7: "Argument 2 must be in the range 1..7";
        require j ge 1 and j le 7: "Argument 3 must be in the range 1..7";
        require i ne j: "Arguments 2 and 3 cannot be equal";
        return i lt j select [A[w[1],w[2]] : w in complementary_entries_7(i,j)]
                        else [A[w[2],w[1]] : w in complementary_entries_7(i,j)];
    end if;
    // Haven't landed in one of the cases we cover, so raise an error.
    require false: "Argument 1 must be a 5x5 or 7x7 matrix";
end intrinsic;

/////////////////////////////////////////////////////////////////////////
// CacheClearFormats
/////////////////////////////////////////////////////////////////////////
// Clears the internal formats caches.
/////////////////////////////////////////////////////////////////////////

intrinsic CacheClearFormats()
{Clear the internal formats cache}
    sing_cache_clear();
    cand_cache_clear();
end intrinsic;
