freeze;

/////////////////////////////////////////////////////////////////////////
// utilities.m
/////////////////////////////////////////////////////////////////////////
// Helpful functions when writing formats.
/////////////////////////////////////////////////////////////////////////
// Authors: Gavin Brown and Alexander Kasprzyk.
/////////////////////////////////////////////////////////////////////////

// Validates the input values. Returns true followed by the validated and
// correctly typed input values on success, false followed by an error message
// otherwise.
function validate_search_input(start,finish,minweight,singtype,
										file,latex,filterF,filterX)
	// Sanity check
    bool,start := IsCoercible(Integers(),start);
    if not bool then
		return false,"Parameter 'start' must be an integer",_,_,_,_,_,_,_;
	end if;
    bool,finish := IsCoercible(Integers(),finish);
    if not bool then
		return false,"Parameter 'finish' must be an integer",_,_,_,_,_,_,_;
	end if;
	if start gt finish then
		return false,"The start value must be less than or equal to the finish value",_,_,_,_,_,_,_;
	end if;
    bool,minweight := IsCoercible(Integers(),minweight);
    if not bool then
		return false,"Parameter 'minweight' must be an integer",_,_,_,_,_,_,_;
	end if;
	if not Type(singtype) eq MonStgElt then
		return false,"Parameter 'singtype' must be a string",_,_,_,_,_,_,_;
	end if;
    if not "isolated" in Split(singtype,",") then
		return false,"Parameter 'singtype' must include 'isolated' - until non-isolated orbifold Riemann-Roch formula exists",_,_,_,_,_,_,_;
    end if;
	if not (file cmpeq false or Type(file) eq MonStgElt) then
		return false,"Parameter 'file' must be a string",_,_,_,_,_,_,_;
	end if;
	if not (latex cmpeq false or Type(latex) eq MonStgElt) then
		return false,"Parameter 'latex' must be a string",_,_,_,_,_,_,_;
	end if;
	if not (filterF cmpeq false or ISA(Type(filterF),Program)) then
		return false,"Parameter 'filterF' must be a function",_,_,_,_,_,_,_;
	end if;
	if not (filterX cmpeq false or ISA(Type(filterX),Program)) then
		return false,"Parameter 'filterX' must be a function",_,_,_,_,_,_,_;
	end if;
	// If necessary, open the output file
	file := file cmpeq false or file eq "" select false else Open(file,"w");
	// If necessary, open the LaTeX output file
    latex := latex cmpeq false or latex eq "" select false else Open(latex,"w");
	// Looks good
	return true,start,finish,minweight,singtype,file,latex,filterF,filterX;
end function;

// Attempts to construct all varieties with given canonical class k and
// dimension d using the formats produced by 'list_formats', with parameters in
// the range 'start' to 'finish' (inclusive). Returns true and the list of
// candidate varieties on success, false followed  by an error message
// otherwise.
// Note: We like to imagine that the parameter is the adjunction degree of the
// format, although this is not enforced in any way.
function search_format( d, k, list_formats :
	start := 0,
	finish := 20,
	minweight := 1,
	singtype := "isolated,quotient,canonical",
	file := false,
	latex := false,
	filterF := false,
	filterX := false )
	bool,start,finish,minweight,singtype,file,latex,filterF,filterX:=
		validate_search_input(start,finish,minweight,singtype,
							  file,latex,filterF,filterX);
	if not bool then return false,start; end if;
	results := [];
    for parameter in [start..finish] do
        vprint Format: "------------------",parameter;
        old_number := #results;
        tt := Cputime();
        for F in list_formats(parameter : filterF:=filterF ) do
            Xs:=CandidatesFromFormat(F,d,k :
			        singtype:=singtype, minweight:=minweight, filter:=filterX);
            results cat:= Xs;
            if Type(file) ne BoolElt then
                for X in Xs do
                    Put(file,Trim(CandidateToString(X)) cat "\n" cat
                             Trim(FormatToString(Format(X))) cat "\n\n");
                end for;
            end if;
			if Type(latex) ne BoolElt then
				for X in Xs do
					Put(latex,CandidateToString(X : level:="Latex"));
				end for;
			end if;
        end for;
        vprintf Format: "New: %o\tTotal: %o\n",#results-old_number,#results;
        vprint Format: "Time:",Cputime(tt);
	end for;
    return true,results;
end function;
